SBSMSLogo ICON "sbsms.ico"
/////////////////////////////////////////////////////////////////////////////
//
// Version
//
VS_VERSION_INFO VERSIONINFO
 FILEVERSION 1,7,0,0
 PRODUCTVERSION 1,7,0,0
 FILEFLAGSMASK 0x17L
#ifdef _DEBUG
 FILEFLAGS 0x1L
#else
 FILEFLAGS 0x0L
#endif
 FILEOS 0x4L
 FILETYPE 0x1L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0"
        BEGIN
            VALUE "CompanyName", "Mune"
            VALUE "FileDescription", "SBSMS wxPlayer"
            VALUE "FileVersion", "1.7.0"
            VALUE "InternalName", "SBSMS"
            VALUE "LegalCopyright", "Copyright © 2008. All rights reserved."
            VALUE "OriginalFilename", "sbsms.exe"
            VALUE "ProductName", "SBSMS wxPlayer"
            VALUE "ProductVersion", "1.7.0"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1200
    END
END
